import 'package:masamune/masamune.dart';
import 'package:masamune_module/variable/date_time_form_config_builder.dart';
import 'package:masamune_module/variable/hidden_form_config_builder.dart';
import 'package:masamune_module/variable/multiple_select_form_config_builder.dart';
import 'package:masamune_module/variable/multiple_text_form_config_builder.dart';
import 'package:masamune_module/variable/select_form_config_builder.dart';
import 'package:masamune_module/variable/slider_form_config_builder.dart';
import 'package:masamune_module/variable/text_form_config_builder.dart';

const _variables = <FormConfigBuilder>[
  TextFormConfigBuilder(),
  SelectFormConfigBuilder(),
  DateTimeFormConfigBuilder(),
  MultipleSelectFormConfigBuilder(),
  MultipleTextFormConfigBuilder(),
  SliderFormConfigBuilder(),
  HiddenFormConfigBuilder(),
];

class VariableConfigDefinition {
  const VariableConfigDefinition._();

  /// VariableConfig definition of the name.
  static const VariableConfig name = VariableConfig(
    id: Const.name,
    label: "Name",
    required: true,
    form: TextFormConfig(),
  );

  /// VariableConfig definition of the text.
  static const VariableConfig text = VariableConfig(
    id: Const.text,
    label: "Text",
    form: TextFormConfig(maxLines: 5, minLines: 5),
  );

  /// VariableConfig definition of the gender.
  static const VariableConfig gender = VariableConfig(
    id: "gender",
    label: "gender",
    form: SelectFormConfig(
      items: {
        "male": "Male",
        "female": "Female",
        "other": "Others",
      },
      initialKey: "other",
    ),
  );

  /// VariableConfig definition of the age.
  static const VariableConfig ages = VariableConfig(
    id: "ages",
    label: "Ages",
    form: SelectFormConfig(
      items: {
        "teens": "10s",
        "twenties": "20s",
        "thirties": "30s",
        "forties": "40s",
        "fifties": "50s",
        "sixties": "60s",
      },
      initialKey: "twenties",
    ),
  );

  /// VariableConfig definition of the order.
  static const VariableConfig order = VariableConfig(
    id: "order",
    label: "Order",
    form: HiddenFormConfig(
      type: HiddenFormConfigType.initialOrder,
      applyOnUpdate: false,
    ),
  );

  /// VariableConfig definition of the created time.
  static const VariableConfig createdTime = VariableConfig(
    id: "createdTime",
    label: "Created time",
    form: HiddenFormConfig(
      type: HiddenFormConfigType.dateTimeNow,
      applyOnUpdate: false,
    ),
  );

  /// VariableConfig definition of the updated time.
  static const VariableConfig updatedTime = VariableConfig(
    id: "updatedTime",
    label: "Updated time",
    form: HiddenFormConfig(
      type: HiddenFormConfigType.dateTimeNow,
    ),
  );
}

@immutable
abstract class FormConfigBuilder {
  const FormConfigBuilder();
  bool check(FormConfig? form);

  Iterable<Widget> view(
    VariableConfig config,
    FormConfig? form,
    BuildContext context,
    WidgetRef ref, {
    DynamicMap? data,
    bool onlyRequired = false,
  });

  Iterable<Widget> form(
    VariableConfig config,
    FormConfig? form,
    BuildContext context,
    WidgetRef ref, {
    DynamicMap? data,
    bool onlyRequired = false,
  });

  dynamic value(
    VariableConfig config,
    BuildContext context,
    WidgetRef ref,
    bool updated,
  );
}

extension VariableConfigBuildExtensions on VariableConfig {
  Iterable<Widget> buildView(
    BuildContext context,
    WidgetRef ref, {
    DynamicMap? data,
    bool onlyRequired = false,
  }) {
    if (onlyRequired && !this.required) {
      return const [];
    }
    if (!show) {
      return const [];
    }
    for (final variable in _variables) {
      if (!variable.check(form)) {
        continue;
      }
      return variable.view(
        this,
        form,
        context,
        ref,
        data: data,
        onlyRequired: onlyRequired,
      );
    }
    return const [];
  }

  Iterable<Widget> buildForm(
    BuildContext context,
    WidgetRef ref, {
    DynamicMap? data,
    bool onlyRequired = false,
  }) {
    if (onlyRequired && !this.required) {
      return const [];
    }
    for (final variable in _variables) {
      if (!variable.check(form)) {
        continue;
      }
      return variable.form(
        this,
        form,
        context,
        ref,
        data: data,
        onlyRequired: onlyRequired,
      );
    }
    return const [];
  }

  dynamic _buildValue(
    BuildContext context,
    WidgetRef ref, {
    bool updated = true,
  }) {
    for (final variable in _variables) {
      if (!variable.check(form)) {
        continue;
      }
      return variable.value(
        this,
        context,
        ref,
        updated,
      );
    }
    return null;
  }

  dynamic buildValue(
    DynamicMap data,
    BuildContext context,
    WidgetRef ref, {
    bool onlyRequired = false,
    bool updated = true,
  }) {
    if (onlyRequired && !this.required) {
      return;
    }
    final value = _buildValue(
      context,
      ref,
      updated: updated,
    );
    if (value == null) {
      return;
    }
    data[id] = value;
  }
}

extension VariableConfigListExtensions on Iterable<VariableConfig>? {
  Iterable<Widget> buildView(
    BuildContext context,
    WidgetRef ref, {
    DynamicMap? data,
    bool onlyRequired = false,
  }) {
    if (this == null) {
      return const [];
    }
    return this!.expand((e) =>
        e.buildView(context, ref, data: data, onlyRequired: onlyRequired));
  }

  Iterable<Widget> buildForm(
    BuildContext context,
    WidgetRef ref, {
    DynamicMap? data,
    bool onlyRequired = false,
  }) {
    if (this == null) {
      return const [];
    }
    return this!.expand((e) =>
        e.buildForm(context, ref, data: data, onlyRequired: onlyRequired));
  }

  Map<String, dynamic> buildMap(
    BuildContext context,
    WidgetRef ref, {
    bool onlyRequired = false,
  }) {
    if (this == null) {
      return const {};
    }
    return this!
        .where((e) => !onlyRequired || e.required)
        .toMap<String, dynamic>(
          key: (val) => val.id,
          value: (val) => val._buildValue(
            context,
            ref,
            updated: false,
          ),
        );
  }

  void buildValue(
    DynamicMap data,
    BuildContext context,
    WidgetRef ref, {
    bool onlyRequired = false,
    bool updated = true,
  }) {
    if (this == null) {
      return;
    }
    for (final value in this!) {
      value.buildValue(
        data,
        context,
        ref,
        onlyRequired: onlyRequired,
        updated: updated,
      );
    }
  }
}
